package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ByteArraySshjFile;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.BestNodeLoginFinder;
import be.iminds.ilabt.jfed.util.KeyUtil;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.sftp.*;
import net.schmizz.sshj.transport.TransportException;
import net.schmizz.sshj.transport.verification.PromiscuousVerifier;
import net.schmizz.sshj.userauth.UserAuthException;
import net.schmizz.sshj.userauth.keyprovider.KeyPairWrapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.net.SocketFactory;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

/**
 * ExperimentSpecificationServicesDirsAndUploadsState does 3 things:
 *   - Setup SSH connections to nodes, for the dirs, upload, and execute steps
 *   - Do the "dir" step of the ExperimentSpecification
 *   - Do the "upload" step of the ExperimentSpecification
 *
 * Note that these 2 steps can be combined here easily, as there are no inter node dependencies.
 * For the execute step, things are different. See ExperimentSpecificationServicesSingleExecuteState
 *
 * Additionally, this class can be called to close all SSH connections. (typically after all executes)
 *   See TODO
 */
public class ExperimentSpecificationServicesDirsAndUploadsState extends SlicedState<ExperimentSpecificationServicesDirsAndUploadsState.ExperimentSpecificationServiceDirsAndUploadsStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesDirsAndUploadsState.class);
    @Nonnull
    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    @Nonnull
    private final Experiment experiment;

    @Nonnull
    private final ExperimentSpecification experimentSpecification;
    @Nonnull
    private final ESpecBundle experimentSpecificationBundle;
    @Nonnull
    private final ESpecLogger eSpecLogger;


    @Nonnull
    private final List<ExperimentSpecificationServiceDirsAndUploadsStateSlice> slices;

    @Nonnull
    private final ConcurrentHashMap<UploadLikeSpec, String> remotePathByUploadLikeSpec = new ConcurrentHashMap<>();

    protected ExperimentSpecificationServicesDirsAndUploadsState(@Nonnull SetupSoftwareExperimentJob job) {
        super("Execute ExperimentSpecification Dir And Upload Steps");
        this.setupSoftwareExperimentJob = job;
        this.experiment = job.getExperiment();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();

        experiment.requireSlice();
        if (experiment.getExperimentSpecification() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification");
        if (experiment.getExperimentSpecificationBundle() == null)
            throw new IllegalStateException("Internal error, ExperimentSpecificationServicesDirsAndUploadsState created for experiment without ExperimentSpecification (bundle)");

        this.experimentSpecification = experiment.getExperimentSpecification();
        this.experimentSpecificationBundle = experiment.getExperimentSpecificationBundle();

        assert experiment.getKeypairs() != null;

        this.slices = new ArrayList<>();


        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            LOG.warn("ExperimentSpecificationServicesDirsAndUploadsState will do nothing because there is no usable request RSpec");
            return;
        }

        for (FXRspecNode node : model.getNodes()) {
            String nodeId = node.getClientId();

            if (nodeId != null) {
                //Note: Since this needs to setup SSH connections for execute as well, this will also run on nodes with only an execute step.

                boolean neededOnNode = experimentSpecification.getDirs().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getExecutes().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));
                neededOnNode |= experimentSpecification.getUploads().stream().anyMatch(s -> s.getNodes() == null || s.getNodes().contains(nodeId));

                if (neededOnNode) {
                    slices.add(new ExperimentSpecificationServiceDirsAndUploadsStateSlice(job, node));
                }
            }
        }
    }

    @Nonnull
    @Override
    protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
        ExperimentTaskStatus res = super.executeState(job);
        eSpecLogger.firePreExecute();
        return res;
    }


    @Override
    public Collection<ExperimentSpecificationServiceDirsAndUploadsStateSlice> getSlices() {
        return slices;
    }

    @Nullable
    public String getRemotePath(@Nonnull UploadLikeSpec uploadLikeSpec) {
        return remotePathByUploadLikeSpec.get(uploadLikeSpec);
    }


    public class AssureAllSshClosedStateSlice extends State {
        protected AssureAllSshClosedStateSlice() {
            super("Close All SSH Connections");
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            setupSoftwareExperimentJob.closeAllSshConnections();
            return ExperimentTaskStatus.SUCCESS;
        }
    }

    public AssureAllSshClosedStateSlice getAssureAllSshClosedStateSlice() {
        return new AssureAllSshClosedStateSlice();
    }

    public class ExperimentSpecificationServiceDirsAndUploadsStateSlice extends StateSlice {
        @Nonnull
        protected final FXRspecNode node;
        @Nonnull
        protected final String clientId;
        @Nullable
        private SSHClient ssh;

        @Nullable
        private ExternalFileUtil externalFileUtil;

        public ExperimentSpecificationServiceDirsAndUploadsStateSlice(@Nonnull Job<?> job,
                                                                      @Nonnull FXRspecNode node) {
            super(job);
            this.node = node;
            if (node.getClientId() == null)
                throw new IllegalStateException("Only nodes with a client_id can be used in an ExperimentSpecification");
            this.clientId = node.getClientId();
        }


        private String dirPathToAbsolutePath(String dirPath) {
            if (externalFileUtil == null)
                throw new IllegalStateException("externalFileUtil == null");

            //3 cases for path:
            //  - absolute path (starts with /)   example: /work/data/ / /work /work/test/data
            //  - home path (starts with ~)       exammple: ~/data/  ~/   ~   ~/test/data
            //  - null -> home path               -> this is handled by the parser, so no need to check this here
            //TODO    empty path string or path not starting with / or ~  ->  error during parsing

            boolean absolutePath = dirPath.startsWith("/");
            assert absolutePath || dirPath.startsWith("~");

            if (absolutePath)
                return assureEndWithSlash(dirPath);

            String homeDir = assureEndWithSlash(externalFileUtil.getHomeDir());

            return assureEndWithSlash(homeDir + (dirPath.length() > 1 ? dirPath.substring(1) : ""));
        }

        @Override
        public String getName() {
            return "ExperimentSpecification steps for node " + node.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(node, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExperimentSpecificationServiceDirsAndUploadsStateSlice.statefulRun due to SSH comnection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            externalFileUtil = new ExternalFileUtil(ssh);

            eSpecLogger.firePreDir();
            //TODO: only start when needed
            final SetupDirsState setupDirState = new SetupDirsState();
            setAndRunState(setupDirState);
            eSpecLogger.firePostDirAll(setupDirState.getStatus() == ExperimentTaskStatus.SUCCESS);

            if (setupDirState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;


            eSpecLogger.firePreUpload();
            final UploadState uploadState = new UploadState();
            try {
                setAndRunState(uploadState);
            } finally {
                eSpecLogger.firePostUploadAll(uploadState.getStatus() == ExperimentTaskStatus.SUCCESS);
            }

            if (uploadState.getStatus() != ExperimentTaskStatus.SUCCESS)
                return ExperimentTaskStatus.FAILED;

            return ExperimentTaskStatus.SUCCESS;
        }

        public class SetupDirsState extends State {
            protected SetupDirsState() {
                super("Checking (and possible creating) dirs on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                List<DirSpec> dirs = new ArrayList<>(experimentSpecification.getDirs());

                for (DirSpec dirSpec : dirs) {
                    if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                        String remotePath = dirPathToAbsolutePath(dirSpec.getPath());

                        int targetDirMask = ChModPermissionsUtil.toMask(dirSpec.getPermissions());
                        FileAttributes targetDirAtts = new FileAttributes.Builder()
                                .withPermissions(targetDirMask)
                                .withType(FileMode.Type.DIRECTORY).build();

                        try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                            FileAttributes att = sftpClient.statExistence(remotePath);

                            LOG.debug("ExperimentSpecification.dir checked dir {} on {}. exists={}",
                                    remotePath, node.getClientId(), att != null);

                            if (att != null) {
                                //handle existing dir
                                if (att.getType() != FileMode.Type.DIRECTORY) {
                                    throw new IllegalStateException("Error while checking dir "+remotePath+", it already exists but is not a directory");
                                }

                                //don't check permissions, just set them
                                sftpClient.chmod(remotePath, targetDirMask);
                            } else {
                                //create dir(s)
                                SFTPEngine engine = sftpClient.getSFTPEngine();
                                final Deque<String> dirsToMake = new LinkedList<>();
                                for (PathComponents current = engine.getPathHelper().getComponents(remotePath); ;
                                     current = engine.getPathHelper().getComponents(current.getParent())) {
                                    final FileAttributes attrs = sftpClient.statExistence(current.getPath());
                                    if (attrs == null) {
                                        dirsToMake.push(current.getPath());
                                    } else if (attrs.getType() != FileMode.Type.DIRECTORY) {
                                        throw new IllegalStateException("Error while creating dir "+remotePath+" -> "+current.getPath() + " exists but is not a directory");
                                    } else {
                                        break;
                                    }
                                }
                                LOG.debug("ExperimentSpecification.dir must create dirs {}", dirsToMake);
                                while (!dirsToMake.isEmpty()) {
                                    String dirToMake = dirsToMake.pop();
                                    LOG.debug("ExperimentSpecification.dir creating {} on {}",
                                            dirToMake, node.getClientId());
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    engine.makeDir(dirToMake, targetDirAtts);
                                    //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                    sftpClient.chmod(dirToMake, targetDirMask);
                                }
                            }

                            //TODO: we should set appropriate permissions along the parts of the directory path that already exist. ("execute" perm for uog depending if they get any permissions on final dir)

                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, true);
                        } catch (IOException ex) {
                            LOG.error("Error while creating dir {} on {}", remotePath, node.getClientId(), ex);
                            updateMessage(String.format("Error while creating dir %s on %s: %s", remotePath, node.getClientId(), ex.getMessage()));
                            eSpecLogger.firePostDir(dirSpec, remotePath, clientId, false);
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class UploadState extends State {
            protected UploadState() {
                super("Uploading files on " + node.getClientId());
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                if (ssh == null)
                    throw new IllegalStateException("ssh == null");
                if (externalFileUtil == null)
                    throw new IllegalStateException("externalFileUtil == null");

                //We need to wait for all files to be ready
                //noinspection ConstantConditions
                job.getExperiment().getExperimentSpecificationFileManager().waitUntilReady();

                //uploads file AND scripts (but don't execute scripts yet)
                List<UploadLikeSpec> uploads = new ArrayList<>(experimentSpecification.getUploads());
                uploads.addAll(experimentSpecification.getExecutes());

                String uploadDirSpecPath = null;
                String scriptDirSpecPath = null;
                for (DirSpec dirSpec : experimentSpecification.getDirs()) {
                    if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                        if (dirSpec.isScriptContent()) {
                            scriptDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                        if (dirSpec.isUploadContent()) {
                            uploadDirSpecPath = dirPathToAbsolutePath(dirSpec.getPath());
                        }
                    }
                }

                try (SFTPClient sftpClient = ssh.newSFTPClient()) {

                    for (UploadLikeSpec uploadSpec : uploads) {
                        if (uploadSpec.getNodes() == null || uploadSpec.getNodes().contains(clientId)) {
                            byte[] content;
                            if (uploadSpec.getSource().isGeneratedKeyPair()) {
                                content = getGeneratedContent(uploadSpec.getSource().getValue());
                            } else {
                                content = job.getExperiment().getExperimentSpecificationFileManager().getBytes(uploadSpec.getSource());
                            }

                            assert externalFileUtil != null;
                            //use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
                            String remotePath;
                            if (uploadSpec.getPath() == null) {
                                if (uploadSpec instanceof UploadSpec) {
                                    if (uploadDirSpecPath != null) {
                                        remotePath = assureEndWithSlash(convertHomeDir(uploadDirSpecPath, externalFileUtil.getHomeDir())) +
                                                uploadSpec.getSource().getBasename();
                                    } else {
                                        remotePath = assureEndWithSlash(externalFileUtil.getHomeDir()) + uploadSpec.getSource().getBasename();
                                    }
                                } else {
                                    if (uploadSpec instanceof ExecuteSpec) {
                                        if (scriptDirSpecPath != null) {
                                            remotePath = assureEndWithSlash(convertHomeDir(scriptDirSpecPath, externalFileUtil.getHomeDir())) +
                                                    uploadSpec.getSource().getBasename();
                                        } else {
                                            remotePath = assureEndWithSlash(externalFileUtil.getHomeDir()) + uploadSpec.getSource().getBasename();
                                        }
                                    } else {
                                        throw new RuntimeException("BUG: uploadSpec of unknown class " + uploadSpec.getClass().getName());
                                    }
                                }
                            } else {
                                remotePath = convertHomeDir(uploadSpec.getPath(), externalFileUtil.getHomeDir());
                            }

                            String name = externalFileUtil.getFilename(remotePath);

                            int targetFileMask = ChModPermissionsUtil.toMask(uploadSpec.getPermissions());

                            try {
                                LOG.debug("ExperimentSpecification.upload is uploading {} ({} bytes) to {} on {}",
                                        name, content.length, remotePath, node.getClientId());
                                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                sftpClient.getFileTransfer().upload(
                                        new ByteArraySshjFile(name, content, uploadSpec.getPermissions()),
                                        remotePath);
                                //TODO   NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work...
                                sftpClient.chmod(remotePath, targetFileMask);

                                remotePathByUploadLikeSpec.put(uploadSpec, remotePath);

                                eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, true);
                            } catch (IOException ex) {
                                LOG.error("Error while uploading {} to {} on {}", name, remotePath, node.getClientId(), ex);
                                updateMessage(String.format("Error while uploading %s to %s on %s: %s", name, remotePath, node.getClientId(), ex.getMessage()));
                                eSpecLogger.firePostUpload(uploadSpec, remotePath, clientId, false);
                                return ExperimentTaskStatus.FAILED;
                            }
                        }
                    }
                } catch (IOException e) {
                    throw new JFedException("Failed to create SSH SFTP client", e);
                }

                return ExperimentTaskStatus.SUCCESS;
            }

            private byte[] getGeneratedContent(String generatedContentId) {
                switch (generatedContentId) {
                    case "keypair.private": {
                        if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                            throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                        KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                        assert keyPair != null;
                        assert keyPair.getPrivate() != null;
                        char[] privateKeyChars = KeyUtil.privateKeyToPem(keyPair.getPrivate());
                        return new String(privateKeyChars).getBytes(StandardCharsets.UTF_8);
                    }
                    case "keypair.public": {
                        if (experiment.getKeypairs() == null || experiment.getKeypairs().isEmpty())
                            throw new IllegalStateException("\"generated\" source \"" + generatedContentId + "\" used, but experiment has no keypairs");
                        KeyPair keyPair = experiment.getKeypairs().values().iterator().next();
                        assert keyPair != null;
                        assert keyPair.getPublic() != null;
                        String publicKeyString = KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(keyPair.getPublic());
                        return publicKeyString.getBytes(StandardCharsets.UTF_8);
                    }
                    default: throw new UnsupportedOperationException("Unsupported \"generated\" source \""+generatedContentId+"\"");
                }
            }
        }
    }

    @Nonnull
    private static String convertHomeDir(@Nonnull String path, @Nonnull String homeDirPath) {
        if (path.startsWith("~/") && path.length() > 2) return assureEndWithSlash(homeDirPath)+path.substring(2);
        if (path.startsWith("~") && path.length() > 1) return assureEndWithSlash(homeDirPath)+path.substring(1);
        if (path.equals("~")) return assureEndWithSlash(homeDirPath);
        return path;
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
